@@ -851,12 +851,213 @@ static void polymod_mul_si(element_t e, element_t f, signed long int z)
     }
 }
 
-static void polymod_mul(element_ptr res, element_ptr e, element_ptr f)
+//Karatsuba for degree 2 polynomials
+static void kar_poly_2(element_t *dst, element_t c3, element_t c4, element_t *s1, element_t *s2, element_t *scratch)
+{
+    element_ptr c01, c02, c12;
+
+    c12 = scratch[0];
+    c02 = scratch[1];
+    c01 = scratch[2];
+
+    element_add(c3, s1[0], s1[1]);
+    element_add(c4, s2[0], s2[1]);
+    element_mul(c01, c3, c4);
+    element_add(c3, s1[0], s1[2]);
+    element_add(c4, s2[0], s2[2]);
+    element_mul(c02, c3, c4);
+    element_add(c3, s1[1], s1[2]);
+    element_add(c4, s2[1], s2[2]);
+    element_mul(c12, c3, c4);
+
+    element_mul(dst[1], s1[1], s2[1]);
+
+    //constant term
+    element_mul(dst[0], s1[0], s2[0]);
+
+    //coefficient of x^4
+    element_mul(c4, s1[2], s2[2]);
+
+    //coefficient of x^3
+    element_add(c3, dst[1], c4);
+    element_sub(c3, c12, c3);
+
+    //coefficient of x^2
+    element_add(dst[2], c4, dst[0]);
+    element_sub(c02, c02, dst[2]);
+    element_add(dst[2], dst[1], c02);
+
+    //coefficient of x
+    element_sub(c01, c01, dst[0]);
+    element_sub(dst[1], c01, dst[1]);
+}
+
+static void polymod_mul_degree3(element_ptr res, element_ptr e, element_ptr f)
 {
     polymod_field_data_ptr p = res->field->data;
-    element_t *dst, *s1 = e->data, *s2 = f->data;
+    element_t *dst = res->data, *s1 = e->data, *s2 = f->data;
+    element_t c3, c4;
+    element_t p0;
+
+    element_init(p0, res->field);
+    element_init(c3, p->field);
+    element_init(c4, p->field);
+
+    kar_poly_2(dst, c3, c4, s1, s2, p0->data);
+
+    polymod_const_mul(p0, c3, p->xpwr[0]);
+    element_add(res, res, p0);
+    polymod_const_mul(p0, c4, p->xpwr[1]);
+    element_add(res, res, p0);
+
+    element_clear(p0);
+    element_clear(c3);
+    element_clear(c4);
+}
+
+static void polymod_mul_degree6(element_ptr res, element_ptr e, element_ptr f)
+{
+    polymod_field_data_ptr p = res->field->data;
+    element_t *dst = res->data, *s0, *s1 = e->data, *s2 = f->data;
+    element_t *a0, *a1, *b0, *b1;
+    element_t p0, p1, p2, p3;
+
+    a0 = s1;
+    a1 = &s1[3];
+    b0 = s2;
+    b1 = &s2[3];
+
+    element_init(p0, res->field);
+    element_init(p1, res->field);
+    element_init(p2, res->field);
+    element_init(p3, res->field);
+
+    s0 = p0->data;
+    s1 = p1->data;
+    s2 = p2->data;
+    element_add(s0[0], a0[0], a1[0]);
+    element_add(s0[1], a0[1], a1[1]);
+    element_add(s0[2], a0[2], a1[2]);
+
+    element_add(s1[0], b0[0], b1[0]);
+    element_add(s1[1], b0[1], b1[1]);
+    element_add(s1[2], b0[2], b1[2]);
+
+    kar_poly_2(s2, s2[3], s2[4], s0, s1, p3->data);
+    kar_poly_2(s0, s0[3], s0[4], a0, b0, p3->data);
+    kar_poly_2(s1, s1[3], s1[4], a1, b1, p3->data);
+
+    element_set(dst[0], s0[0]);
+    element_set(dst[1], s0[1]);
+    element_set(dst[2], s0[2]);
+
+    element_sub(dst[3], s0[3], s0[0]);
+    element_sub(dst[3], dst[3], s1[0]);
+    element_add(dst[3], dst[3], s2[0]);
+
+    element_sub(dst[4], s0[4], s0[1]);
+    element_sub(dst[4], dst[4], s1[1]);
+    element_add(dst[4], dst[4], s2[1]);
+
+    element_sub(dst[5], s2[2], s0[2]);
+    element_sub(dst[5], dst[5], s1[2]);
+
+    //start reusing part of s0 as scratch(!)
+    element_sub(s0[0], s2[3], s0[3]);
+    element_sub(s0[0], s0[0], s1[3]);
+    element_add(s0[0], s0[0], s1[0]);
+
+    element_sub(s0[1], s2[4], s0[4]);
+    element_sub(s0[1], s0[1], s1[4]);
+    element_add(s0[1], s0[1], s1[1]);
+
+    polymod_const_mul(p3, s0[0], p->xpwr[0]);
+    element_add(res, res, p3);
+    polymod_const_mul(p3, s0[1], p->xpwr[1]);
+    element_add(res, res, p3);
+    polymod_const_mul(p3, s1[2], p->xpwr[2]);
+    element_add(res, res, p3);
+    polymod_const_mul(p3, s1[3], p->xpwr[3]);
+    element_add(res, res, p3);
+    polymod_const_mul(p3, s1[4], p->xpwr[4]);
+    element_add(res, res, p3);
+
+    element_clear(p0);
+    element_clear(p1);
+    element_clear(p2);
+    element_clear(p3);
+}
+
+/*
+void polymod_mul_degree3(element_ptr res, element_ptr e, element_ptr f)
+{
+    polymod_field_data_ptr p = res->field->data;
+    element_t *dst = res->data, *s1 = e->data, *s2 = f->data;
+    element_t *p0data;
+    element_ptr c01, c02, c12;
+    element_t c3, c4;
+    element_t p0;
+
+    element_init(p0, res->field);
+    element_init(c3, p->field);
+    element_init(c4, p->field);
+    p0data = p0->data;
+
+    c12 = p0data[0];
+    c02 = p0data[1];
+    c01 = p0data[2];
+
+    element_add(c3, s1[0], s1[1]);
+    element_add(c4, s2[0], s2[1]);
+    element_mul(c01, c3, c4);
+    element_add(c3, s1[0], s1[2]);
+    element_add(c4, s2[0], s2[2]);
+    element_mul(c02, c3, c4);
+    element_add(c3, s1[1], s1[2]);
+    element_add(c4, s2[1], s2[2]);
+    element_mul(c12, c3, c4);
+
+    element_mul(dst[1], s1[1], s2[1]);
+
+    //constant term
+    element_mul(dst[0], s1[0], s2[0]);
+
+    //coefficient of x^4
+    element_mul(c4, s1[2], s2[2]);
+
+    //coefficient of x^3
+    element_add(c3, dst[1], c4);
+    element_sub(c3, c12, c3);
+
+    //coefficient of x^2
+    element_add(dst[2], c4, dst[0]);
+    element_sub(c02, c02, dst[2]);
+    element_add(dst[2], dst[1], c02);
+
+    //coefficient of x
+    element_sub(c01, c01, dst[0]);
+    element_sub(dst[1], c01, dst[1]);
+
+    polymod_const_mul(p0, c3, p->xpwr[0]);
+    element_add(res, res, p0);
+    polymod_const_mul(p0, c4, p->xpwr[1]);
+    element_add(res, res, p0);
+
+    element_clear(p0);
+    element_clear(c3);
+    element_clear(c4);
+}
+*/
+
+static void polymod_mul(element_ptr res, element_ptr e, element_ptr f)
+//NOTE: cases n = 3, 6 have dedicated routines
+{
+    polymod_field_data_ptr p = res->field->data;
+    int n = p->n;
+    element_t *dst;
+    element_t *s1 = e->data, *s2 = f->data;
     element_t prod, p0, c0;
-    int i, j, n = p->n;
+    int i, j;
     element_t *high; //coefficients of x^n,...,x^{2n-2}
 
     high = malloc(sizeof(element_t) * (n - 1));
@@ -894,12 +1095,58 @@ static void polymod_mul(element_ptr res, element_ptr e, element_ptr f)
     element_clear(c0);
 }
 
+static void polymod_square_degree3(element_ptr res, element_ptr e)
+//TODO: investigate if squaring is significantly cheaper
+//than multiplication. If so convert to Karatsuba.
+{
+    element_t *dst = res->data;
+    element_t *src = e->data;
+    polymod_field_data_ptr p = res->field->data;
+    element_t p0;
+    element_t c0, c2;
+    element_ptr c1, c3;
+
+    element_init(p0, res->field);
+    element_init(c0, p->field);
+    element_init(c2, p->field);
+
+    c3 = p0->data;
+    c1 = c3 + 1;
+
+    element_mul(c3, src[0], src[1]);
+    element_mul(c1, src[0], src[2]);
+    element_square(dst[0], src[0]);
+
+    element_mul(c2, src[1], src[2]);
+    element_square(c0, src[2]);
+    element_square(dst[2], src[1]);
+
+    element_add(dst[1], c3, c3);
+
+    element_add(c1, c1, c1);
+    element_add(dst[2], dst[2], c1);
+
+    polymod_const_mul(p0, c0, p->xpwr[1]);
+    element_add(res, res, p0);
+
+    element_add(c2, c2, c2);
+    polymod_const_mul(p0, c2, p->xpwr[0]);
+    element_add(res, res, p0);
+
+    element_clear(p0);
+    element_clear(c0);
+    element_clear(c2);
+}
+
 static void polymod_square(element_ptr res, element_ptr e)
 {
+    element_t *dst;
+    element_t *src = e->data;
     polymod_field_data_ptr p = res->field->data;
-    element_t *dst, *src = e->data;
+    int n = p->n;
+
     element_t prod, p0, c0;
-    int i, j, n = p->n;
+    int i, j;
 
     element_t *high; //coefficients of x^n,...,x^{2n-2}
 
@@ -1050,6 +1297,7 @@ void field_init_polymod(field_ptr f, element_ptr poly)
 {
     poly_field_data_ptr pdp = poly->field->data;
     polymod_field_data_ptr p;
+    int n;
 
     field_init(f);
     f->data = malloc(sizeof(polymod_field_data_t));
@@ -1058,7 +1306,7 @@ void field_init_polymod(field_ptr f, element_ptr poly)
     p->mapbase = element_field_to_poly;
     element_init(p->poly, poly->field);
     element_set(p->poly, poly);
-    p->n = poly_degree(p->poly);
+    n = p->n = poly_degree(p->poly);
     f->field_clear = field_clear_polymod;
     f->init = polymod_init;
     f->clear = polymod_clear;
@@ -1075,10 +1323,23 @@ void field_init_polymod(field_ptr f, element_ptr poly)
     f->set0 = polymod_set0;
     f->set1 = polymod_set1;
     f->cmp = polymod_cmp;
-    f->mul = polymod_mul;
+    switch(n) {
+	case 3:
+	    f->mul = polymod_mul_degree3;
+	    f->square = polymod_square_degree3;
+	    break;
+	case 6:
+	    f->mul = polymod_mul_degree6;
+	    f->square = polymod_square;
+	    break;
+	default:
+	    f->mul = polymod_mul;
+	    f->square = polymod_square;
+	    break;
+    }
+
     f->mul_mpz = polymod_mul_mpz;
     f->mul_si = polymod_mul_si;
-    f->square = polymod_square;
     f->random = polymod_random;
     f->from_hash = polymod_from_hash;
     f->invert = polymod_invert;
@@ -1092,9 +1353,9 @@ void field_init_polymod(field_ptr f, element_ptr poly)
     } else {
 	f->fixed_length_in_bytes = pdp->field->fixed_length_in_bytes * poly_degree(poly);
     }
-    mpz_pow_ui(f->order, p->field->order, p->n);
+    mpz_pow_ui(f->order, p->field->order, n);
 
-    p->xpwr = malloc(sizeof(element_t) * p->n);
+    p->xpwr = malloc(sizeof(element_t) * n);
     compute_x_powers(f, poly);
 }
 
